"""The frame ignoring system for varname

There 4 mechanisms to ignore intermediate frames to determine the desired one
so that a variable name should be retrieved at that frame.

1. Ignore frames by a given module. Any calls inside it and inside its
   submodules will be ignored. A filename (path) to a module is also acceptable
   and recommended when code is executed by `exec` without module available.
2. Ignore frames by a given pair of module and a qualified name (qualname).
   See 1) for acceptable modules. The qualname should be unique in that module.
3. Ignore frames by a (non-decorated) function.
4. Ignore frames by a decorated function. In this case, you can specified a
   tuple with the function and the number of decorators of it. The decorators
   on the wrapper function inside the decorators should also be counted.

Any frames in `varname`, standard libraries, and frames of any expressions like
<lambda> are ignored by default.

"""
import sys
import inspect
import distutils.sysconfig as sysconfig
import warnings
from os import path
from pathlib import Path
from fnmatch import fnmatch
from abc import ABC, abstractmethod
from typing import List, Union
from types import FrameType, ModuleType, FunctionType

from executing import Source

from .utils import (
    IgnoreElemType,
    IgnoreType,
    MaybeDecoratedFunctionWarning,
    cached_getmodule,
    attach_ignore_id_to_module,
    frame_matches_module_by_ignore_id,
    check_qualname_by_source,
    debug_ignore_frame,
)

class IgnoreElem(ABC):
    """An element of the ignore list"""

    def __init_subclass__(cls, attrs: List[str]) -> None:
        """Define different attributes for subclasses"""

        def subclass_init(
            self,
            # IgnoreModule: ModuleType
            # IgnoreFilename/IgnoreDirname: str
            # IgnoreFunction: FunctionType
            # IgnoreDecorated: FunctionType, int
            # IgnoreModuleQualname/IgnoreFilenameQualname:
            #   ModuleType/str, str
            # IgnoreOnlyQualname: None, str
            *ign_args: Union[str, int, ModuleType, FunctionType],
        ) -> None:
            """__init__ function for subclasses"""
            for attr, arg in zip(attrs, ign_args):
                setattr(self, attr, arg)

            self._post_init()

        # save it for __repr__
        cls.attrs = attrs
        cls.__init__ = subclass_init # type: ignore

    def _post_init(self) -> None:
        """Setups after __init__"""

    @abstractmethod
    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        """Whether the frame matches the ignore element"""

    def __repr__(self) -> str:
        """Representation of the element"""
        attr_values = (getattr(self, attr) for attr in self.__class__.attrs)
        # get __name__ if possible
        attr_values = (
            repr(getattr(attr_value, "__name__", attr_value))
            for attr_value in attr_values
        )
        attr_values = ", ".join(attr_values)
        return f"{self.__class__.__name__}({attr_values})"


class IgnoreModule(IgnoreElem, attrs=["module"]):
    """Ignore calls from a module or its submodules"""

    def _post_init(self) -> None:
        attach_ignore_id_to_module(self.module)

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        frame = frameinfos[frame_no].frame
        module = cached_getmodule(frame.f_code)
        if module:
            return (
                module.__name__ == self.module.__name__
                or module.__name__.startswith(f"{self.module.__name__}.")
            )

        return frame_matches_module_by_ignore_id(frame, self.module)


class IgnoreFilename(IgnoreElem, attrs=["filename"]):
    """Ignore calls from a module by matching its filename"""

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        frame = frameinfos[frame_no].frame

        # in case of symbolic links
        return path.realpath(frame.f_code.co_filename) == path.realpath(
            self.filename
        )


class IgnoreDirname(IgnoreElem, attrs=["dirname"]):
    """Ignore calls from modules inside a directory

    Currently used internally to ignore calls from standard libraries."""

    def _post_init(self) -> None:
        # pylint: disable=access-member-before-definition
        # pylint: disable=attribute-defined-outside-init

        # Path object will turn into str here
        self.dirname = path.realpath(self.dirname) # type: str

        if not self.dirname.endswith(path.sep):
            self.dirname = f"{self.dirname}{path.sep}"

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        frame = frameinfos[frame_no].frame
        filename = path.realpath(frame.f_code.co_filename)

        return filename.startswith(self.dirname)


class IgnoreStdlib(IgnoreDirname, attrs=["dirname"]):
    """Ignore standard libraries in sysconfig.get_python_lib(standard_lib=True)

    But we need to ignore 3rd-party packages under site-packages/.
    """

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        frame = frameinfos[frame_no].frame
        third_party_lib = f"{self.dirname}site-packages{path.sep}"
        filename = path.realpath(frame.f_code.co_filename)

        return (
            filename.startswith(self.dirname)
            and
            # Exclude 3rd-party libraries in site-packages
            not filename.startswith(third_party_lib)
        )


class IgnoreFunction(IgnoreElem, attrs=["func"]):
    """Ignore a non-decorated function"""

    def _post_init(self) -> None:
        if (
            # without functools.wraps
            "<locals>" in self.func.__qualname__
            or self.func.__name__ != self.func.__code__.co_name
        ):
            warnings.warn(
                f"You asked varname to ignore function {self.func.__name__!r}, "
                "which may be decorated. If it is not intended, you may need "
                "to ignore all intermediate frames with a tuple of "
                "the function and the number of its decorators.",
                MaybeDecoratedFunctionWarning,
            )

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        frame = frameinfos[frame_no].frame
        return frame.f_code == self.func.__code__


class IgnoreDecorated(IgnoreElem, attrs=["func", "n_decor"]):
    """Ignore a decorated function"""

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        try:
            frame = frameinfos[frame_no + self.n_decor].frame
        except IndexError:
            return False

        return frame.f_code == self.func.__code__


class IgnoreModuleQualname(IgnoreElem, attrs=["module", "qualname"]):
    """Ignore calls by qualified name in the module"""

    def _post_init(self) -> None:

        attach_ignore_id_to_module(self.module)
        # check uniqueness of qualname
        modfile = getattr(self.module, "__file__", None)
        if modfile is not None:
            check_qualname_by_source(
                Source.for_filename(modfile, self.module.__dict__),
                self.module.__name__,
                self.qualname,
            )

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        frame = frameinfos[frame_no].frame
        module = cached_getmodule(frame.f_code)

        # Return earlier to avoid qualname uniqueness check
        if module and module != self.module:
            return False

        if not module and not frame_matches_module_by_ignore_id(
            frame, self.module
        ):
            return False

        source = Source.for_frame(frame)
        check_qualname_by_source(source, self.module.__name__, self.qualname)

        return fnmatch(source.code_qualname(frame.f_code), self.qualname)


class IgnoreFilenameQualname(IgnoreElem, attrs=["filename", "qualname"]):
    """Ignore calls with given qualname in the module with the filename"""

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        frame = frameinfos[frame_no].frame

        frame_filename = path.realpath(frame.f_code.co_filename)
        preset_filename = path.realpath(self.filename)
        # return earlier to avoid qualname uniqueness check
        if frame_filename != preset_filename:
            return False

        source = Source.for_frame(frame)
        check_qualname_by_source(source, self.filename, self.qualname)

        return fnmatch(source.code_qualname(frame.f_code), self.qualname)


class IgnoreOnlyQualname(IgnoreElem, attrs=["_none", "qualname"]):
    """Ignore calls that match the given qualname, across all frames."""

    def match(self, frame_no: int, frameinfos: List[inspect.FrameInfo]) -> bool:
        frame = frameinfos[frame_no].frame

        # module is None, check qualname only
        return fnmatch(
            Source.for_frame(frame).code_qualname(frame.f_code), self.qualname
        )


def create_ignore_elem(ignore_elem: IgnoreElemType) -> IgnoreElem:
    """Create an ignore element according to the type"""
    if isinstance(ignore_elem, ModuleType):
        return IgnoreModule(ignore_elem)  # type: ignore
    if isinstance(ignore_elem, (Path, str)):
        return (
            IgnoreDirname(ignore_elem)  # type: ignore
            if path.isdir(ignore_elem)
            else IgnoreFilename(ignore_elem) # type: ignore
        )
    if hasattr(ignore_elem, "__code__"):
        return IgnoreFunction(ignore_elem) # type: ignore
    if not isinstance(ignore_elem, tuple) or len(ignore_elem) != 2:
        raise ValueError(f"Unexpected ignore item: {ignore_elem!r}")
    # is tuple and len == 2
    if hasattr(ignore_elem[0], "__code__") and isinstance(ignore_elem[1], int):
        return IgnoreDecorated(*ignore_elem) # type: ignore
    # otherwise, the second element should be qualname
    if not isinstance(ignore_elem[1], str):
        raise ValueError(f"Unexpected ignore item: {ignore_elem!r}")

    if isinstance(ignore_elem[0], ModuleType):
        return IgnoreModuleQualname(*ignore_elem) # type: ignore
    if isinstance(ignore_elem[0], (Path, str)):
        return IgnoreFilenameQualname(*ignore_elem) # type: ignore
    if ignore_elem[0] is None:
        return IgnoreOnlyQualname(*ignore_elem)

    raise ValueError(f"Unexpected ignore item: {ignore_elem!r}")


class IgnoreList:
    """The ignore list to match the frames to see if they should be ignored"""

    @classmethod
    def create(
        cls,
        ignore: IgnoreType = None,
        ignore_lambda: bool = True,
        ignore_varname: bool = True,
    ) -> "IgnoreList":
        """Create an IgnoreList object

        Args:
            ignore: An element of the ignore list, either
                A module (or filename of a module)
                A tuple of module (or filename) and qualified name
                A function
                A tuple of function and number of decorators
            ignore_lambda: whether ignore lambda functions
            ignore_varname: whether the calls from this package

        Returns:
            The IgnoreList object
        """
        ignore = ignore or []
        if not isinstance(ignore, list):
            ignore = [ignore]

        ignore_list = [
            IgnoreStdlib( # type: ignore
                sysconfig.get_python_lib(standard_lib=True)
            )
        ] # type: List[IgnoreElem]
        if ignore_varname:
            ignore_list.append(create_ignore_elem(sys.modules[__package__]))
        if ignore_lambda:
            ignore_list.append(create_ignore_elem((None, "*<lambda>")))
        for ignore_elem in ignore:
            ignore_list.append(create_ignore_elem(ignore_elem))

        return cls(ignore_list) # type: ignore

    def __init__(self, ignore_list: List[IgnoreElemType]) -> None:
        self.ignore_list = ignore_list
        debug_ignore_frame(
            '>>> IgnoreList initiated <<<'
        )

    def nextframe_to_check(
        self, frame_no: int, frameinfos: List[inspect.FrameInfo]
    ) -> int:
        """Find the next frame to check

        In modst cases, the next frame to check is the next adjacent frame.
        But for IgnoreDecorated, the next frame to check should be the next
        `ignore[1]`th frame.

        Args:
            frame_no: The index of current frame to check
            frameinfos: The frame info objects

        Returns:
            A number for Next `N`th frame to check. 0 if no frame matched.
        """
        for ignore_elem in self.ignore_list:
            matched = ignore_elem.match(frame_no, frameinfos) # type: ignore
            if matched and isinstance(ignore_elem, IgnoreDecorated):
                debug_ignore_frame(
                    f"Ignored by {ignore_elem!r}", frameinfos[frame_no]
                )
                return ignore_elem.n_decor + 1

            if matched:
                debug_ignore_frame(
                    f"Ignored by {ignore_elem!r}", frameinfos[frame_no]
                )
                return 1
        return 0

    def get_frame(self, frame_no: int) -> FrameType:
        """Get the right frame by the frame number

        Args:
            frame_no: The index of the frame to get

        Returns:
            The desired frame

        Raises:
            VarnameRetrievingError: if any exceptions raised during the process.
        """
        try:
            # since this function will be called by APIs
            # so we should skip that
            frames = inspect.getouterframes(sys._getframe(2), 0)
            i = 0

            while i < len(frames):
                nextframe = self.nextframe_to_check(i, frames)
                # ignored
                if nextframe > 0:
                    i += nextframe
                    continue

                frame_no -= 1
                if frame_no == 0:
                    debug_ignore_frame("Gotcha!", frames[i])
                    return frames[i].frame

                debug_ignore_frame(
                    f"Skipping ({frame_no - 1} more to skip)", frames[i]
                )
                i += 1

        except Exception as exc:
            from .utils import VarnameRetrievingError

            raise VarnameRetrievingError from exc

        return None # pragma: no cover



"""Some internal utilities for varname


Attributes:

    IgnoreElemType: The type for ignore elements
    IgnoreType: The type for the ignore argument
    MODULE_IGNORE_ID_NAME: The name of the ignore id injected to the module.
        Espectially for modules that can't be retrieved by
        `inspect.getmodule(frame)`
"""
import sys
import dis
import ast
import warnings
import inspect
from os import path
from pathlib import Path
from functools import lru_cache
from types import ModuleType, FunctionType, CodeType, FrameType
from typing import Tuple, Union, List, Mapping, Callable

from executing import Source

IgnoreElemType = Union[
    # module
    ModuleType,
    # filename of a module
    str,
    Path,
    FunctionType,
    # the module (filename) and qualname
    # If module is None, then all qualname matches the 2nd element
    # will be ignored. Used to ignore <lambda> internally
    Tuple[Union[ModuleType, str], str],
    # Function and number of its decorators
    Tuple[FunctionType, int],
]
IgnoreType = Union[IgnoreElemType, List[IgnoreElemType]]

ArgSourceType = Union[ast.AST, str]
ArgSourceType = Union[ArgSourceType, Tuple[ArgSourceType, ...]]
ArgSourceType = Union[ArgSourceType, Mapping[str, ArgSourceType]]

MODULE_IGNORE_ID_NAME = "__varname_ignore_id__"


class config:  # pylint: disable=invalid-name
    """Global configurations for varname

    Attributes:
        debug: Show debug information for frames being ignored
    """

    debug = False


class VarnameRetrievingError(Exception):
    """When failed to retrieve the varname"""


class QualnameNonUniqueError(Exception):
    """When a qualified name is used as an ignore element but references to
    multiple objects in a module"""


class NonVariableArgumentError(Exception):
    """When vars_only is True but try to retrieve name of
    a non-variable argument"""


class ImproperUseError(Exception):
    """When varname() is improperly used"""


class MaybeDecoratedFunctionWarning(Warning):
    """When a suspecious decorated function used as ignore function directly"""


class MultiTargetAssignmentWarning(Warning):
    """When varname tries to retrieve variable name in
    a multi-target assignment"""


@lru_cache()
def cached_getmodule(codeobj: CodeType):
    """Cached version of inspect.getmodule"""
    return inspect.getmodule(codeobj)


def get_node(
    frame: int,
    ignore: IgnoreType = None,
    raise_exc: bool = True,
    ignore_lambda: bool = True,
) -> ast.AST:
    """Try to get node from the executing object.

    This can fail when a frame is failed to retrieve.
    One case should be when python code is executed in
    R pacakge `reticulate`, where only first frame is kept.

    When the node can not be retrieved, try to return the first statement.
    """
    from .ignore import IgnoreList

    ignore = IgnoreList.create(ignore, ignore_lambda=ignore_lambda)
    try:
        frameobj = ignore.get_frame(frame)
    except VarnameRetrievingError:
        return None

    return get_node_by_frame(frameobj, raise_exc)


def get_node_by_frame(
    frame: FrameType, raise_exc: bool = True
) -> ast.AST:
    """Get the node by frame, raise errors if possible"""
    exect = Source.executing(frame)

    if exect.node:
        return exect.node

    if exect.source.text and exect.source.tree and raise_exc:
        raise VarnameRetrievingError(
            "Couldn't retrieve the call node. "
            "This may happen if you're using some other AST magic at the "
            "same time, such as pytest, ipython, macropy, or birdseye."
        )

    return None


def lookfor_parent_assign(node: ast.AST) -> Union[ast.Assign, ast.AnnAssign]:
    """Look for an ast.Assign node in the parents"""
    while hasattr(node, "parent"):
        node = node.parent

        if isinstance(node, (ast.AnnAssign, ast.Assign)):
            return node
    return None


def node_name(node: ast.AST) -> Union[str, Tuple[Union[str, Tuple], ...]]:
    """Get the node node name.

    Raises ImproperUseError when failed
    """
    if isinstance(node, ast.Name):
        return node.id
    if isinstance(node, ast.Attribute):
        return node.attr
    if isinstance(node, (ast.List, ast.Tuple)):
        return tuple(node_name(elem) for elem in node.elts)

    raise ImproperUseError(
        f"Can only get name of a variable or attribute, "
        f"not {ast.dump(node)}"
    )


@lru_cache()
def bytecode_nameof(code: CodeType, offset: int) -> str:
    """Cached Bytecode version of nameof

    We are trying this version only when the sourcecode is unavisible. In most
    cases, this will happen when user is trying to run a script in REPL/
    python shell, with `eval`, or other circumstances where the code is
    manipulated to run but sourcecode is not available.
    """
    instructions = list(dis.get_instructions(code))
    ((current_instruction_index, current_instruction),) = (
        (index, instruction)
        for index, instruction in enumerate(instructions)
        if instruction.offset == offset
    )

    if current_instruction.opname in ("CALL_FUNCTION_EX", "CALL_FUNCTION_KW"):
        raise VarnameRetrievingError(
            "'nameof' can only be called with a single positional argument "
            "when source code is not avaiable."
        )

    if current_instruction.opname not in ("CALL_FUNCTION", "CALL_METHOD"):
        raise VarnameRetrievingError("Did you call 'nameof' in a weird way?")

    name_instruction = instructions[current_instruction_index - 1]

    if not name_instruction.opname.startswith("LOAD_"):
        raise VarnameRetrievingError("Argument must be a variable or attribute")

    name = name_instruction.argrepr
    if not name.isidentifier():
        raise VarnameRetrievingError(
            f"Found the variable name {name!r} which is obviously wrong. "
            "This may happen if you're using some other AST magic at the "
            "same time, such as pytest, ipython, macropy, or birdseye."
        )

    return name


def attach_ignore_id_to_module(module: ModuleType) -> None:
    """Attach the ignore id to module

    This is useful when a module cannot be retrieved by frames using
    `inspect.getmodule`, then we can use this id, which will exist in
    `frame.f_globals` to check if the module matches in ignore.

    Do it only when the __file__ is not avaiable or does not exist for
    the module. Since this probably means the source is not avaiable and
    `inspect.getmodule` would not work
    """
    module_file = getattr(module, "__file__", None)
    if module_file is not None and path.isfile(module_file):
        return
    # or it's already been set
    if hasattr(module, MODULE_IGNORE_ID_NAME):
        return

    setattr(module, MODULE_IGNORE_ID_NAME, f"<varname-ignore-{id(module)})")


def frame_matches_module_by_ignore_id(
    frame: FrameType, module: ModuleType
) -> bool:
    """Check if the frame is from the module by ignore id"""
    ignore_id_attached = getattr(module, MODULE_IGNORE_ID_NAME, object())
    ignore_id_from_frame = frame.f_globals.get(MODULE_IGNORE_ID_NAME, object())
    return ignore_id_attached == ignore_id_from_frame


@lru_cache()
def check_qualname_by_source(
    source: Source, modname: str, qualname: str
) -> None:
    """Check if a qualname in module is unique"""
    if not source.tree:
        # no way to check it, skip
        return
    nobj = list(source._qualnames.values()).count(qualname)
    if nobj > 1:
        raise QualnameNonUniqueError(
            f"Qualname {qualname!r} in "
            f"{modname!r} refers to multiple objects."
        )


def debug_ignore_frame(
    msg: str,
    frameinfo: inspect.FrameInfo = None
) -> None:
    """Print the debug message for a given frame info object

    Args:
        msg: The debugging message
        frameinfo: The FrameInfo object for the frame
    """
    if not config.debug:
        return
    if frameinfo is not None:
        msg = (
            f"{msg} [In {frameinfo.function!r} at "
            f"{frameinfo.filename}:{frameinfo.lineno}]"
        )
    sys.stderr.write(f"[{__package__}] DEBUG: {msg}\n")


def argnode_source(
    source: Source, node: ast.AST, vars_only: bool
) -> Union[str, ast.AST]:
    """Get the source of an argument node

    Args:
        source: The executing source object
        node: The node to get the source from
        vars_only: Whether only allow variables and attributes

    Returns:
        The source of the node (node.id for ast.Name,
            node.attr for ast.Attribute). Or the node itself if the source
            cannot be fetched.
    """
    if vars_only:
        return (
            node.id
            if isinstance(node, ast.Name)
            else node.attr
            if isinstance(node, ast.Attribute)
            else node
        )

    # requires asttokens
    return source.asttokens().get_text(node)


@lru_cache()
def get_argument_sources(
    source: Source,
    node: ast.Call,
    func: Callable,
    vars_only: bool,
    pos_only: bool,
) -> Mapping[str, ArgSourceType]:
    """Get the sources for argument from an ast.Call node

    >>> def func(a, b, c, d=4):
    >>>  ...
    >>> x = y = z = 1
    >>> func(y, x, c=z)
    >>> # argument_sources = {'a': 'y', 'b', 'x', 'c': 'z'}
    >>> func(y, x, c=1)
    >>> # argument_sources = {'a': 'y', 'b', 'x', 'c': ast.Num(n=1)}
    """
    # <Signature (a, b, c, d=4)>
    signature = inspect.signature(func, follow_wrapped=False)
    # func(y, x, c=z)
    # ['y', 'x'], {'c': 'z'}
    arg_sources = [
        argnode_source(source, argnode, vars_only) for argnode in node.args
    ]
    kwarg_sources = (
        {
            argnode.arg: argnode_source(source, argnode.value, vars_only)
            for argnode in node.keywords
            if argnode.arg is not None
        }
        if not pos_only
        else {}
    )
    bound_args = signature.bind_partial(*arg_sources, **kwarg_sources)
    argument_sources = bound_args.arguments
    # see if *args and **kwargs have anything assigned
    # if not, assign () and {} to them
    for parameter in signature.parameters.values():
        if parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            argument_sources.setdefault(parameter.name, ())
        if parameter.kind == inspect.Parameter.VAR_KEYWORD:
            argument_sources.setdefault(parameter.name, {})
    return argument_sources


def get_function_called_argname(frame: FrameType, node: ast.AST) -> Callable:
    """Get the function who called argname"""
    # We need node to be ast.Call
    if not isinstance(node, ast.Call):
        raise VarnameRetrievingError(
            f"Expect an 'ast.Call' node, but got {type(node)!r}. "
            "Are you using 'argname' inside a function?"
        )

    # variable
    if isinstance(node.func, ast.Name):
        func = frame.f_locals.get(
            node.func.id, frame.f_globals.get(node.func.id)
        )
        if func is None:  # pragma: no cover
            # not sure how it would happen but in case
            raise VarnameRetrievingError(
                f"Cannot retrieve the function by {node.func.id!r}."
            )
        return func

    # use pure_eval
    pure_eval_fail_msg = None
    try:
        from pure_eval import Evaluator, CannotEval
    except ImportError:
        pure_eval_fail_msg = "'pure_eval' is not installed."
    else:
        try:
            evaluator = Evaluator.from_frame(frame)
            return evaluator[node.func]
        except CannotEval:
            pure_eval_fail_msg = (
                f"Cannot evaluate node {ast.dump(node.func)} "
                "using 'pure_eval'."
            )

    # try eval
    warnings.warn(
        f"{pure_eval_fail_msg} "
        "Using 'eval' to get the function that calls 'argname'. "
        "Try calling it using a variable reference to the function, or "
        "passing the function to 'argname' explicitly."
    )
    expr = ast.Expression(node.func)
    code = compile(expr, "<ast-call>", "eval")
    # pylint: disable=eval-used
    return eval(code, frame.f_globals, frame.f_locals)


def parse_argname_subscript(node: ast.Subscript):
    """Parse the ast.Subscript node passed to argname

    Make sure:
    1. node.value is an ast.Name node
    2. node.slice is a ast.Constant node

    This is separated as a function, since we want to see in the future whether
    non-ast.Constant node can be supported as node.slice.
    """
    name = node.value
    if not isinstance(name, ast.Name):
        raise ValueError(f"Expect {ast.dump(name)} to be a variable.")

    subscript = node.slice # type: ast.AST
    if isinstance(subscript, ast.Index):
        subscript = subscript.value
    if not isinstance(subscript, (ast.Str, ast.Num, ast.Constant)):
        raise ValueError(f"Expect {ast.dump(subscript)} to be a constant.")

    subscript = getattr(
        subscript, "value", subscript
    )  # ast.Index, ast.Constant
    subscript = getattr(subscript, "n", subscript)  # ast.Num
    subscript = getattr(subscript, "s", subscript)  # ast.Str

    return name.id, subscript
